#____________________________________________________________________________________
#
#    PersistentThreads (see README.txt for more information)
#    Copyright (C) 2002 Christoph Becker-Freyseng
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#    christoph@mmc-startup.com
#    http://www.beyond-thoughts.com
#    http://www.mmc-startup.com
#____________________________________________________________________________________
#



__doc__="""PPManager is your Interface to PersistentThreads. With an instance of
	    PPManager your can add, del, start, stop and retrieve status of 
	    a PersistentProcessor.
	    You can get an PPManager-Instance by calling _getPPMan(zObj), where 
	    zObj is a Zope-Object which is part of the Acquisition- and 
	    Persistence-Tree. Or you can add the PPManUtilMixin to your
	    Zope-Classes, so you can just say self._getPPMan 
	    (then your Zope-Object will be used as zObj automagically)"""
__version__='0.1.1'



from PersistentDict import PersistentDict
from PConstants import product_id, persistent_threads_id
import PThreadMan
import PPHandler

from App.Common import aq_base

import threading






class PPManager:
    _lock= threading.Lock()
    
    def __init__(self, prod):
	self.prod= prod
	PPManager._lock.acquire()
	try:
	    if hasattr(PPManager,'_init'):
		PPManager._init(self)
		delattr(PPManager,'_init')
	except:
	    import traceback, sys
	    sys.stderr.write('During Initialization of PersistentThreads the following error occurred:\n')
	    traceback.print_exc(file=sys.stderr)
	PPManager._lock.release()
    
    def _init(self):
	PPHandler.initPPHandler(self.prod)	# Make sure that it exists
	PPManager.threadMan= PThreadMan.PThreadMan()
	for (ptID, pp) in self._getPPDict().items():
	    if pp._mature:
		self.threadMan.addPP(ptID, pp.user, pp.passwd)
		self.threadMan.startPP(ptID)	# Just use normal delay Zope is not really running at this moment.   Here it's OK to start them immediately, cause they are alreay in ZODB
    
    def _getPPDict(self):
	try:
	    ppDict= getattr(self.prod, persistent_threads_id, None)
	except:
	    ppDict= None
##	ppDict= None	# Force a new ppDict (Debugging)
	if ppDict is None:
	    ppDict= PersistentDict()    # not a normal dictionary! Important for Persistence AND Acquisition
	    setattr(self.prod, persistent_threads_id, ppDict)
	    self.prod._p_changed= 1
	    get_transaction().note('Added a "'+str(persistent_threads_id)+'" to the "'+str(self.prod)+'" Product')
	    get_transaction().commit()
	ppDict= ppDict.__of__(self.prod)
	return ppDict
    
    def addPP(self, ptID, ppP):  # return 1 if succesful or 0 if failed
	ptID= str(ptID)	# Only Strings are valid as Identifiers
	dictD= self._getPPDict()
	if dictD.has_key(ptID):
	    return 0
	dictD[ptID]= ppP
	self.threadMan.addPP(ptID, ppP.user, ppP.passwd)
	return 1
    
    def getPP(self, ptID):
	return self._getPPDict().get(ptID, None)
    
    def manipulatePP(self, ptID, manipulatorP):
	ppD= self._getPPDict().get(ptID, None)
	if ppD is not None:
	    return self.threadMan.manipulatePP(ptID, ppD.user, ppD.passwd, manipulatorP)
	else:	# did not exist
	    return None
    
    def delPP(self, ptID):
	dictD= self._getPPDict()
	if dictD.has_key(ptID):
	    del dictD[ptID]
	    return 1
	else:
	    return 0
    
    def startPP(self, ptID, delayP=10.3):  # return 1 if succesful or 0 if failed
	ppD= self._getPPDict().get(ptID, None)
	if ppD is not None:
	    if not ppD._mature:
		ppD._mature= 1		# Here there are no conflict-problems, because it surely isn't and wasn't running now
	    if not self.threadMan.hasPP(ptID):	# this is a zombie, wake him
		self.threadMan.addPP(ptID, ppD.user, ppD.passwd)
	    return self.threadMan.startPP(ptID, delayP)
	else:
	    return 0
    
    def stopPP(self, ptID):  # return 1 if succesful or 0 if failed, BLOCKS until pThread is stopped!
	if self.threadMan.hasPP(ptID):
	    return self.threadMan.stopPP(ptID)
	else:	# did not exist/run
	    return None
    
    def statusPP(self, ptID):	# status: -666 means Zombie, None=does not exist, -1=dying, 0=running, >0 means running with this waitTime
	dictD= self._getPPDict()
	if dictD.has_key(ptID):
	    return self.threadMan.statusPP(ptID)
	else:	# did not exist
	    return None






def _getPPMan(zObj):
    if hasattr(zObj, 'getPhysicalRoot'):
	app= zObj.getPhysicalRoot()
    else:
	app= None
	objR= zObj
	while objR:
	    if hasattr(aq_base(objR), 'isTopLevelPrincipiaApplicationObject'):
		if objR.isTopLevelPrincipiaApplicationObject:	# Maybe use objR.aq_base to not acquire isTopLevelPrincipiaApplicationObject
		    app= objR
		    break
	    if hasattr(objR, 'aq_parent'):
		objR= objR.aq_parent
	    else:
		break
    prodD= getattr(app.Control_Panel.Products, product_id)	# If we can't find our product we have an unresolve-able problem -> crash OK
    return PPManager(prodD)

class PPManUtilMixin:
    _getPPMan= _getPPMan







